
/**********************************************************************
 *           gamefield.cpp  -  description 
 *           Project: "Labyrinth"
 *
 *           (C) 2003 by Helge Plehn
 *           29.03.2003
*                                                                         
 *   This program is free software; you can redistribute it and/or modify  
 *   it under the terms of the GNU General Public License as published by 
 *   the Free Software Foundation; either version 2 of the License, or     
 *   (at your option) any later version.                               
 *                                                                    
**********************************************************************/

#include "gamefield.h"
#include "qembed_images.h"



// Just call the generated function
// "name" is the original image filename WITHOUT the extension
QImage myFindImage(const char* name)
{
    return qembed_findImage(name);
}




GameField::GameField( QWidget *parent, const char *name )
: QWidget( parent, name )
{
 GameMode = 0;
 BallMode = 0;


 //loadbitmaps

 bool ok = true;
    
 if ( ok ) ok = ballimage.convertFromImage(myFindImage("kugel"),0);
 if ( ok ) ok = ballmask.convertFromImage(myFindImage("kugel_maske"),0);
 if ( ok ) ok = labyrinthimage.convertFromImage(myFindImage("labyrinth"),0);
    
 if ( !ok )
 {
   //Image could not be loaded
	  QMessageBox::critical ( 0, tr("Error!"), tr("Couldn't access Bitmaps!") );
    close();
 }

 ballcopy = QPixmap( 8, 8, -1 );   //Bits per Pixel = -1 sets the current Bit Depth of the system
 
 // Optimize Speed for Bitblt
 ballimage.setOptimization ( QPixmap::BestOptim ) ;
 ballmask.setOptimization ( QPixmap::BestOptim ) ;
 ballcopy.setOptimization ( QPixmap::BestOptim ) ;
 labyrinthimage.setOptimization ( QPixmap::BestOptim ) ;
 
 //Set the masks for the sprites
 ballimage.setMask( ballmask );

 //gameimage = labyrinthimage;
 gameimage = QPixmap(238,230,-1);
 QPainter p( &gameimage );
 p.drawPixmap( 0,23, labyrinthimage );
 
 
 setBackgroundPixmap( gameimage );      // set desktop pixmap


 // Defines all the Data-Arrays
 DefineArrays();

 // setMouseTracking ( true );  // Enables MouseMoveEvents without any Mousebutton pressed
 setCursor(QCursor(10)); // No Mousecursor
  

 CheatMode = false;
 //Friction = true;
 MouseOverWidget = false;

}


void GameField::DefineArrays()
{
//Array-Werte zuweisen
  const int HolesX[40] = {17, 32, 15, 49, 50, 50, 31, 17, 32, 51, 70, 70, 69, 88,107, 89, 69, 68,107,126,126,128,147,147,148,166,149,147,147,183,202,218,183,203,184,219,184,204,219,204};
  const int HolesY[40] = {19, 43,108, 43, 91,126,146,160,176,167,187,141,104,139,175, 71, 67, 41, 71, 34,103,153, 17, 53, 77, 93,139,168,189, 17, 18, 30, 50, 80, 89,128,148,148,162,187};

  // Coordinates of alls Obstacles {X1,Y1,X2,Y2}
  const int ObstacleCoordinates[56][4] = {{0,0,1520,58},{1466,0,1520,1335},{0,0,58,1335},{0,1274,1520,1335},{0,151,157,189},{362,0,396,134},{0,375,170,413},{372,1197,410,1335},{874,0,903,634},
                            {1229,0,1263,141},{1344,0,1379,138},{1352,735,1520,768},{126,260,164,298},{236,156,274,237},{130,501,275,534},{237,306,275,534},{358,229,396,309},{362,394,396,472},
                            {136,625,170,751},{249,638,286,749},{363,616,398,770},{485,614,519,768},{126,857,281,889},{242,857,281,1182},{373,1013,408,1115},{494,132,528,283},{494,132,903,170},
                            {490,376,525,491},{490,376,645,412},{609,241,645,412},{742,248,774,327},{736,405,771,505},{620,516,652,841},{620,807,769,841},{738,605,772,715},{490,947,635,984},
                            {490,947,524,1172},{614,1068,654,1184},{743,953,778,1188},{866,714,900,946},{869,1126,1022,1162},{1232,1102,1267,1182},{874,225,1001,264},{874,399,1134,433},
                            {1339,234,1381,346},{1105,228,1142,334},{1105,228,1270,266},{1237,228,1270,578},{1352,444,1388,768},{1005,633,1150,668},{1113,528,1150,761},{1241,694,1276,1009},
                            {992,748,1027,909},{992,875,1147,909},{1115,875,1147,1161}};


  // Initiate Holes
  for ( int i = 0; i < 40 ; i++ )
  {
    Holes[i] = QPoint(64*HolesX[i],64*HolesY[i]);    
  };
  // Initiate Obstacles
  for ( int i = 0 ; i < 56 ; i++ )
  {
      Obstacles[i] = MyBlock( QPoint(10*ObstacleCoordinates[i][0],10*ObstacleCoordinates[i][1]),QPoint(10*ObstacleCoordinates[i][2],10*ObstacleCoordinates[i][3]));
  };
  
 
}



void GameField::StartGame()
{
 QPainter p( &gameimage );


 if ( GameMode == 2 )	//game is already running
  return;

 if (GameMode == 1)	  //resume from pause
 {
  GameMode = 2;
  emit setgamemode(2);
  startTimer( 20 );  // Alle 20 Millisekunden = 50 Hz
  return;
 };

  // Start new game

 //gameimage = labyrinthimage;
 gameimage = QPixmap(238,230,-1);
 p.drawPixmap( 0,23, labyrinthimage );
        
 GameMode=2;
 BallMode = 1;
 GameTime=0;
 emit setgamemode(2);
 BallPosition = QPoint(8384,1024);
 BallSpeed = QPoint(0,0);
 TiltVector = QPoint(0,0);
 emit TiltVectorChange( TiltVector);
 SpeedAdd = QPoint(0,0);
 CursorVector=QPoint(0,0);
 startTimer( 20 );   // Alle 20 Millisekunden = 50 Hz
}


void GameField::StopGame()
{
 killTimers();
 GameMode = 0;
 emit setgamemode(0);
 	// set gamefield to start (refresh!)
}


void GameField::PauseGame()
{
 if ( GameMode < 2 )	//game is not running
  return;
 killTimers();
 GameMode = 1;
 emit setgamemode(1);
}


void GameField::spacepressed()
{
 if ( TiltVector == QPoint(0,0) && ( GameMode == 0 || GameMode == 1 ) ) {
   StartGame();
   return;
 }; 
 TiltVector = QPoint(0,0);
 emit TiltVectorChange( TiltVector );  
}




void GameField::mousePressEvent( QMouseEvent *e )
{
  //qDebug( " mousePressEvent ");
  Mausabfrage( e );
}

  
void GameField::mouseMoveEvent( QMouseEvent *e )
{
  //qDebug( " mouseMoveEvent ");
  Mausabfrage( e );
}


void GameField::mouseReleaseEvent(  QMouseEvent *e )
{
 //qDebug( " mouseReleaseEvent ");
 MouseOverWidget = false;
 TiltVector = QPoint(0,0);
 emit TiltVectorChange( TiltVector );     
} 


void GameField::Mausabfrage( QMouseEvent *e )
{
 QPoint TempMousePosition;
 bool TempMouseOverWidget = true;

 TempMousePosition = e -> pos();
 TempMousePosition.setX(TempMousePosition.x() );   //-8
 TempMousePosition.setY(TempMousePosition.y() -23); //-6

 if ( TempMousePosition.x() < 4 || TempMousePosition.x() > 236 || TempMousePosition.y() < 4 || TempMousePosition.y() > 204 ) TempMouseOverWidget = false;

 MousePosition = TempMousePosition;
 MouseOverWidget = TempMouseOverWidget;

 if ( MouseOverWidget ) {
    TiltVector = CalcTiltVector(MousePosition - BallPosition/64);
    emit TiltVectorChange(TiltVector);
 };

}


QPoint GameField::CalcTiltVector(QPoint TestVector)
{
 // Ermittelt den Tiltvektor in Abhngigkeit von der Ballposition und der Mausposition

 static QPoint LastTiltVector;
 static long LastRadius;
 static long LastRadiusQuadrat;
 long Radius;
 long RadiusQuadrat;
 long RadiusIteration;
 int i = 0;
 
 
 if ( LastTiltVector == TestVector ) return TestVector;
 
 RadiusQuadrat = TestVector.x()*TestVector.x()+TestVector.y()*TestVector.y();
 if ( RadiusQuadrat > 576 )
 {
   if ( RadiusQuadrat == LastRadiusQuadrat ) {
     Radius = LastRadius;     
   }
   else {
     if (LastRadius == 0 ) Radius = RadiusQuadrat/24;  // geschtzter Startwert
     else Radius = LastRadius; // Letzter Wert als Startwert der Iteration
     do {
       RadiusIteration = Radius;
       Radius = ( RadiusIteration + RadiusQuadrat/RadiusIteration ) >> 1; //Newton-Verfahren
       i++;
     } while ( QABS(RadiusIteration - Radius)>1 && i<20 );  // Maximal 20 Iterationen sollten gengen!
   };
   
   TestVector = 24*TestVector / static_cast<int>(Radius);
   LastRadius = Radius;   
 };

 LastRadiusQuadrat = RadiusQuadrat;
 LastTiltVector = TestVector;
 return TestVector;

}


void GameField::SetTiltVector( QPoint Vector )
{
  // erhlt den vermutlichen Tiltvector von Showvector.
  // Korrigiert ggfs. den Tiltvector und liefert das Ergebnis zurck an Showvector.

  
  TiltVector = CalcTiltVector(Vector);
  emit TiltVectorChange(TiltVector);
}


void GameField::SetCursorVector( QPoint Vector )
{
  CursorVector = Vector;
}

/*
void GameField::ChangeFriction( bool NewFriction )
{
  Friction = NewFriction;
}
*/


void GameField::ChangeCheatMode( bool NewCheatMode )
{
  CheatMode = NewCheatMode;
}


void GameField::timerEvent( QTimerEvent * )
{
 // is called in normal GameMode (2) at 50Hz

 QPoint NewBallPosition;
 QPoint OldBallPosition;
 QPoint NewBallPosRot;
 QPoint OldBallPosRot;
 
 QPoint Difference;
 unsigned long Abstand;
 
 MyLine MotionVector;
 //MyLine Sekante;
 
 bool CheckComplete;

 MyLine AntwortContainer;

 QPoint Schnittpunkt;
 
 unsigned long MinAbstandSchnittpunkt;
 unsigned long AbstandSchnittpunkt;

 int SchnittpunktKantenIndex = 0;
 int SchnittpunktKante = 0;
 int SchnittpunktBlockIndex = 0;
 
 static int CursorDelay = 0;
 
 
 QPainter p( &gameimage );




  if( GameMode != 2 )
 	{
	 killTimers(); 	
 	 return;
	};


 
 
 // Check Goal
 if ( BallPosition.x() > 13952 && BallPosition.y() > 6656 && BallPosition.y() < 7104 )
 {
   // Victory!
   GameMode = 0;
   emit setgamemode(-2);
   
 };

 
 // Check Holes
 if ( !CheatMode ) {
   for ( int i = 0; i < 40 ; i++ )
   {
     Difference = Holes[i] - BallPosition;
     if (QABS(Difference.x()) < 448 && QABS(Difference.y()) < 448 ) {
       // critical position, check that!
       Abstand = Obstacles[0].PunktAbstand( Holes[i] , BallPosition );
       if ( Abstand < 200704 ) {
         // Nahe am Loch
         SpeedAdd = SpeedAdd + Difference/4;
       };
       if ( Abstand < 102400 ) {
         // Fast im Loch
         SpeedAdd = SpeedAdd + Difference/2;
       };
       if ( Abstand < 36864 ) {
         // Game Over!
         // qDebug( " Ball ist in Loch Nr. " + Obstacles[SchnittpunktBlockIndex].StringFromNumber(i) + " gefallen, Game Over! " );
         GameMode=0;
         emit setgamemode(-1);

         TiltVector = QPoint(0,0);
         emit TiltVectorChange(TiltVector);
         BallSpeed = QPoint(0,0);
         emit SpeedVectorChange(BallSpeed);

         gameimage = QPixmap(238,230,-1);
         p.drawPixmap( 0,23, labyrinthimage );
         setBackgroundPixmap( gameimage );

       };
     };
   };
 };
 

  if( GameMode != 2 )
  {
    killTimers();
    return;
  };


  // calc new ball position
  OldBallPosition = BallPosition;
  NewBallPosition = BallPosition + BallSpeed;


  if ( BallSpeed != QPoint(0,0) ) {
    MotionVector = MyLine(BallPosition,NewBallPosition);  //???

    // check collision
    do {
      MinAbstandSchnittpunkt = 500000000;
      CheckComplete = true;

      NewBallPosRot = MotionVector.Rotate45(NewBallPosition,1);
      OldBallPosRot = MotionVector.Rotate45(OldBallPosition,1);

      for ( int i = 0 ; i < 56 ; i++ )      // Anzahl der Hindernisse == 56
      {
         AntwortContainer = Obstacles[i].Schnittpunkt(NewBallPosition,OldBallPosition,NewBallPosRot,OldBallPosRot);
         if ( AntwortContainer.GetP1() != QPoint(0,0) ) {
            AbstandSchnittpunkt =  Obstacles[i].PunktAbstand( AntwortContainer.GetP1() , OldBallPosition );
            if ( AbstandSchnittpunkt < MinAbstandSchnittpunkt ) {
               CheckComplete = false;
               MinAbstandSchnittpunkt = AbstandSchnittpunkt;
               Schnittpunkt = AntwortContainer.GetP1();
               SchnittpunktKante = AntwortContainer.GetP2().x();
               SchnittpunktKantenIndex = AntwortContainer.GetP2().y();
               SchnittpunktBlockIndex = i;
            };
         };
      };

      if ( !CheckComplete ) {
        // Reflektion
        //AntwortContainer = Obstacles[SchnittpunktBlockIndex].SpeedReflection( BallSpeed, SpeedAdd,SchnittpunktKante, SchnittpunktKantenIndex);
        //BallSpeed = AntwortContainer.GetP1();
        //SpeedAdd = AntwortContainer.GetP2();

        BallSpeed = Obstacles[SchnittpunktBlockIndex].SpeedReflection( BallSpeed, SchnittpunktKante, SchnittpunktKantenIndex, 16);
        SpeedAdd = Obstacles[SchnittpunktBlockIndex].SpeedReflection( SpeedAdd,SchnittpunktKante, SchnittpunktKantenIndex, 4);

        NewBallPosition = Obstacles[SchnittpunktBlockIndex].Reflection(NewBallPosition, Schnittpunkt, SchnittpunktKante, SchnittpunktKantenIndex);
        OldBallPosition = Schnittpunkt;

        if ( SchnittpunktKante == 1 ) {
          // qDebug( " Reflektion an Hindernis " + Obstacles[SchnittpunktBlockIndex].StringFromNumber(SchnittpunktBlockIndex) + " , Kante " + Obstacles[SchnittpunktBlockIndex].StringFromNumber(SchnittpunktKantenIndex) );
        }
        else
        {
          // qDebug( " Reflektion an Hindernis " + Obstacles[SchnittpunktBlockIndex].StringFromNumber(SchnittpunktBlockIndex) + " , Ecke " + Obstacles[SchnittpunktBlockIndex].StringFromNumber(SchnittpunktKantenIndex) );
        };
      };

    } while ( !CheckComplete );
  
 };
 
 // calc new TiltVector
 if ( CursorVector != QPoint(0,0) )
 {
   CursorDelay--;
   if ( CursorDelay <= 0 ) {
     CursorDelay = 5;
     TiltVector = CalcTiltVector(TiltVector + CursorVector);
     emit TiltVectorChange( TiltVector );
   };
 };

 if ( MouseOverWidget ) {
    TiltVector = CalcTiltVector(MousePosition - BallPosition/64);
    emit TiltVectorChange(TiltVector);
 };  

 // Friction
 //if ( Friction ) {
   SpeedAdd.setX( SpeedAdd.x() - Signum( BallSpeed.x() ) );
   SpeedAdd.setY( SpeedAdd.y() - Signum( BallSpeed.y() ) );   
 //};
 
 emit SpeedVectorChange(BallSpeed);

 // calc new ball speed
 SpeedAdd = SpeedAdd + TiltVector;
 BallSpeed = BallSpeed + SpeedAdd/8;
 SpeedAdd = SpeedAdd - 8*(SpeedAdd/8);

 

 
 // set new ball position
 BallPosition = NewBallPosition;


 // calc image
 MoveSprites();

 // set time
 GameTime = GameTime + 1;
 emit SetTime(GameTime);

}




void GameField::MoveSprites()
{
 //Bild neu erstellen
 static QPoint OldBallPosition(-1000,-1000);
 //bool Ball(false);
 int intBallPosX;
 int intBallPosY;
 int intBallPosX_old;
 int intBallPosY_old;

 intBallPosX = BallPosition.x()/64-4;
 intBallPosY = BallPosition.y()/64-4 +23; // 23 = Verschiebung des Bildes nach unten
 intBallPosX_old = OldBallPosition.x()/64-4 ;
 intBallPosY_old = OldBallPosition.y()/64-4 +23;  // 23 = Verschiebung des Bildes nach unten

 OldBallPosition = BallPosition;

 QPainter p( &gameimage );
    
 
 if( GameMode == 2) //Ball = true;
  {
    //Ausblenden des Balles
    bitBlt( &gameimage, intBallPosX_old, intBallPosY_old, &ballcopy, 0, 0, 8, 8, CopyROP, true );
    // Kopieren des Hintergrunds, zu dem sich der Ball bewegt
    bitBlt( &ballcopy, 0, 0, &gameimage, intBallPosX, intBallPosY, 8, 8, CopyROP, true );
    // Ball zeichnen
    bitBlt( &gameimage, intBallPosX, intBallPosY, &ballimage, 0, 0, 8, 8, CopyROP , false );
  }
  else
  {
    //Game Over
    //gameimage = labyrinthimage;
    gameimage = QPixmap(238,230,-1);
    p.drawPixmap( 0,23, labyrinthimage );
 
    //setBackgroundPixmap( gameimage );
    OldBallPosition = QPoint(-10000,-10000);
  };

 setBackgroundPixmap( gameimage );

}




int GameField::Signum( int Argument )
{
  return (Argument < 0) ? -1 : (0 < Argument) ? 1 : 0 ; 
}


/*
void GameField::paintEvent( QPaintEvent * )
{
  QPainter p( this );

  //p.begin ( &gameimage , this);
  p.setPen( SolidLine );

  p.setBrush( black );
  
  for ( int i = 0 ; i < 56 ; i++ )
  {
     p.drawLine( Obstacles[i].MinX()/64+4, Obstacles[i].MinY()/64 +23 ,Obstacles[i].MaxX()/64-4, Obstacles[i].MinY()/64 +23 );
     p.drawLine( Obstacles[i].MinX()/64+4, Obstacles[i].MaxY()/64 +23 ,Obstacles[i].MaxX()/64-4, Obstacles[i].MaxY()/64 +23 );
     p.drawLine( Obstacles[i].MinX()/64, Obstacles[i].MinY()/64+4 +23 ,Obstacles[i].MinX()/64, Obstacles[i].MaxY()/64-4 +23 );
     p.drawLine( Obstacles[i].MaxX()/64, Obstacles[i].MinY()/64+4 +23 ,Obstacles[i].MaxX()/64, Obstacles[i].MaxY()/64-4 +23 );

     p.drawLine( Obstacles[i].MaxX()/64-4, Obstacles[i].MinY()/64 +23 ,Obstacles[i].MaxX()/64, Obstacles[i].MinY()/64+4 +23 );
     p.drawLine( Obstacles[i].MinX()/64+4, Obstacles[i].MinY()/64 +23 ,Obstacles[i].MinX()/64, Obstacles[i].MinY()/64+4 +23 );

     p.drawLine( Obstacles[i].MaxX()/64-4, Obstacles[i].MaxY()/64 +23 ,Obstacles[i].MaxX()/64, Obstacles[i].MaxY()/64-4 +23 );
     p.drawLine( Obstacles[i].MinX()/64+4, Obstacles[i].MaxY()/64 +23 ,Obstacles[i].MinX()/64, Obstacles[i].MaxY()/64-4 +23 );

  };
} */
















                                                                                                                                                

