/**********************************************************************
 *           labyrinth.cpp  -  description
 *           Project: "Labyrinth"
 *
 *           (C) 2003 by Helge Plehn
 *           29.03.2003
 *                                                                         
 *   This program is free software; you can redistribute it and/or modify  
 *   it under the terms of the GNU General Public License as published by 
 *   the Free Software Foundation; either version 2 of the License, or     
 *   (at your option) any later version.                               
 *                                                                    
**********************************************************************/

#include "labyrinth.h"
#include <qpe/qpeapplication.h>


MyWidget::MyWidget( QWidget *parent, const char *name, WFlags fl )
: QWidget( parent, name, fl )
{
 // define widget
 setMinimumSize( 240, 286 );
 setMaximumSize( 240, 286 );

 setCaption( tr("Labyrinth") );
 setBackgroundColor( QColor(0, 0, 0) );

 LCDtime = new QLCDNumber(11 , this, "time"  );
 LCDtime->setGeometry(3, 238, 80, 15 );
 LCDtime->setBackgroundColor( QColor(0, 0, 0) );
 LCDtime->setFrameStyle ( 0 );
 LCDtime->setPalette ( QPalette ( QColor ( 0, 0, 255 ) ) );

 LCDtimerecord = new QLCDNumber(11 , this, "timerecord"  );
 LCDtimerecord->setGeometry( 3, 258, 80, 15 );
 LCDtimerecord->setFrameStyle ( 0 );
 LCDtimerecord->setBackgroundColor( QColor(0, 0, 0) );
 LCDtimerecord->setPalette ( QPalette ( QColor ( 255, 0, 0 ) ) );
 LCDtimerecord->hide();

 
 // Extra Wisgets
 gameField = new GameField( this, "gameField" );
 gameField->setGeometry( 0, 0, 238, 230 );
 
 showVector = new ShowVector( this, "showVector" );
 showVector->setGeometry( 88, 230, 64, 56 );
 
 // define Menu

 QPopupMenu *game = new QPopupMenu();
 CHECK_PTR( game );
 game->insertItem( tr("&Start"), this, SLOT(StartGame()) );
 game->insertItem( tr("&Pause"), gameField, SLOT(PauseGame()) );
 game->insertItem( tr("Stop"), gameField, SLOT(StopGame()) );
 game->insertSeparator();
 game->insertItem( tr("E&xit"),  this , SLOT(close()) );                   // eigentlich nicht "this",  sondern "qApp"

 options = new QPopupMenu();
 CHECK_PTR( options );
 //FrictionID = options->insertItem( " &Friction", this, SLOT(ChangeFriction()) );
 CheatingID = options->insertItem( tr(" &Cheat mode"), this, SLOT(ChangeCheatMode()) );
 options->insertSeparator();
 options->insertItem( tr(" &Delete record"), this, SLOT(DeleteRecord()) );
 options->setCheckable( TRUE );
 //options->setItemChecked( FrictionID, true );
 options->setItemChecked( CheatingID, false );


 QPopupMenu *help = new QPopupMenu();
 CHECK_PTR( help );
 help->insertItem( tr("&Help"), this, SLOT(help()));
 help->insertItem( tr("&About"), this, SLOT(about()) );

 
 menu = new QMenuBar( this );
 CHECK_PTR( menu );
 menu->insertItem( tr("&Game"), game );
 menu->insertItem( tr("&Options"), options );
 menu->insertSeparator();
 menu->insertItem( tr("&Help"), help );
 menu->setSeparator( QMenuBar::InWindowsStyle );

 // define PushButton

 StartPause = new QPushButton(tr("Start"),this);
 StartPause->setGeometry( 160, 240, 70, 37 );
 StartPause->setDefault( false );   // funktioniert nicht??!?
 StartPause->setFocusPolicy(NoFocus);

 // define signals and slots
 // Start/Pause-PushButton -> this
 connect( StartPause ,SIGNAL( clicked() ), this ,SLOT(PushStartPause()));
 // this -> gamefield
 connect( this ,SIGNAL(SpacePressed()), gameField ,SLOT(spacepressed()));
 connect( this ,SIGNAL(EmitStartGame()), gameField ,SLOT(StartGame()));
 connect( this ,SIGNAL(EmitPauseGame()), gameField ,SLOT(PauseGame()));
 connect( this ,SIGNAL(CursorVectorChange(QPoint)), gameField ,SLOT(SetCursorVector(QPoint)));
 connect( this ,SIGNAL(SignalChangeCheatMode(bool)), gameField ,SLOT(ChangeCheatMode(bool)));
// connect( this ,SIGNAL(SignalChangeFriction(bool)), gameField ,SLOT(ChangeFriction(bool)));
 // gamefield -> this
 connect( gameField ,SIGNAL(SetTime(long)), this ,SLOT(SetTime(long)));
 connect( gameField ,SIGNAL(setgamemode(int)), this ,SLOT(setgamemode(int)));
 // gamefield -> showvector
 connect( gameField ,SIGNAL(TiltVectorChange(QPoint)), showVector ,SLOT(SetTiltVector(QPoint)));
 connect( gameField ,SIGNAL(SpeedVectorChange(QPoint)), showVector ,SLOT(SetSpeedVector(QPoint)));
 // showvector -> gamefield
 connect( showVector ,SIGNAL(TiltVectorChange(QPoint)), gameField ,SLOT(SetTiltVector(QPoint)));
 
 cursorup = false;
 cursordown = false;
 cursorleft = false;
 cursorright = false;
 lastleftright = 0;
 lastupdown = 0;
 
 gamemode = 0;
 timerecord = ReadRecordFile();
 if ( timerecord < 17999999 ) LCDtimerecord->show();    
 SetTime(0);

 CheatMode = false;
 //Friction = true;
}


void MyWidget::about()
{
 QMessageBox::about( this, tr("About Labyrinth"),
                         tr(" Labyrinth 0.7 \n\n"
                         " (c) 2003 by Helge Plehn\n"
                         " Helge_Plehn@web.de\n"
                         " This game is free \n software (GPL v2).\n")  );
}


void MyWidget::help()
{
 int answer;
  QMessageBox mb( tr("Labyrinth") , tr(" Use pen or arrowkeys to tilt \n the labyrinth. The tilt vector and \n the speed vector is indicated below. \n The ball will be accelereted \n according to the tilt vector. \n Beware of the inertia! ")
 , QMessageBox::NoIcon, QMessageBox::Ok ,QMessageBox::NoButton , QMessageBox::NoButton , this );
  answer = mb.exec();
}

void MyWidget::StartGame()
{
  cursorup = false;
  cursordown = false;
  cursorleft = false;
  cursorright = false;
  lastleftright = 0;
  lastupdown = 0;

  StartPause->setText(tr("Pause"));

  emit EmitStartGame();
}

void MyWidget::PushStartPause()
{
  if (gamemode == 0 || gamemode == 1) {
    StartGame();
  }
  else {
    emit EmitPauseGame();
  };
}
   
void MyWidget::SetTime(long time)
{
 if (time > 17999999 ) time = 17999999; // mehr als 99h 59min 59,98sec nicht mglich!
 gametime = time;
 LCDtime->display( ConvertTime( time ) );
 LCDtimerecord->display( ConvertTime( timerecord ) );

}


QString MyWidget::ConvertTime( long LongTime )
{
 long Hours;
 long Minutes;
 long Seconds;
 long Hundredths;
 QString StrTime;
 QString StrColon;
 QString StrDigit;
 
 StrColon = ":";
 StrDigit = ".";
 
 
 Hours = LongTime/180000;
 LongTime = LongTime - Hours*180000;
 Minutes = LongTime / 3000;
 LongTime = LongTime - Minutes*3000;
 Seconds = LongTime / 50;
 Hundredths = 2*(LongTime - Seconds*50);

 StrTime = ConvertInt(Hours) + StrColon + ConvertInt(Minutes) + StrColon + ConvertInt(Seconds) + StrDigit + ConvertInt(Hundredths);

 return StrTime;

}


QString MyWidget::ConvertInt(long Nummer )
{
 // Ermittelt einen String aus der Zahl 
 long Zehner;
 long Einer;
 QString StrZehner;
 QString StrEiner;
 
 Zehner = Nummer/10;
 Einer = Nummer - Zehner*10;

 StrZehner.setNum(static_cast<int>(Zehner));
 StrEiner.setNum(static_cast<int>(Einer));
 
 return StrZehner + StrEiner;
 
 //return QString(QChar(static_cast<char>(Zehner))) & QString(QChar(static_cast<char>(Einer)));
 
}



void MyWidget::setgamemode(int Modus)
{
 int answer;

 if( Modus == -1 )
 {
   //Game Over
  
  QMessageBox mb( tr("Labyrinth") , tr("Game over!") , QMessageBox::NoIcon, QMessageBox::Ok ,QMessageBox::NoButton , QMessageBox::NoButton , this );
  answer = mb.exec();
  Modus = 0;
 };

 if( Modus == -2 )
 {
   //Won!
  if (gametime < timerecord && !CheatMode)
  {
    // New Record!
    timerecord = gametime;
    LCDtimerecord->show();
    WriteRecordFile( timerecord );
    SetTime(gametime);
    QMessageBox mb( tr("Labyrinth") , tr("Congratulation, new record!") , QMessageBox::NoIcon, QMessageBox::Ok ,QMessageBox::NoButton , QMessageBox::NoButton , this );
    answer = mb.exec();
  }
  else
  {
    QMessageBox mb( tr("Labyrinth") , tr("You've won, congratulation!") , QMessageBox::NoIcon, QMessageBox::Ok ,QMessageBox::NoButton , QMessageBox::NoButton , this );
    answer = mb.exec();
  }          
  Modus = 0;
 };

 if ( Modus == 2 ) {
   StartPause->setText(tr("Pause"));
 }
 else {
   StartPause->setText(tr("Start"));
 };

 gamemode = Modus;
}

/*
void MyWidget::ChangeFriction()
{
  Friction = !Friction;
  options->setItemChecked( FrictionID, Friction );
  emit SignalChangeFriction(Friction);
  // qDebug( " Friction changed " );
      
}*/


void MyWidget::ChangeCheatMode()
{
  if ( CheatMode && gamemode != 0 ) return;    //Whrend des Spiel kann der Cheatmode nicht abgestellt werden!
  CheatMode = !CheatMode;
  options->setItemChecked( CheatingID, CheatMode );
  emit SignalChangeCheatMode(CheatMode);
  // qDebug( " CheatMode changed " );
      
}

void MyWidget::DeleteRecord()
{
  WriteRecordFile( 17999999 );
  LCDtimerecord->hide();    
};


void MyWidget::keyPressEvent( QKeyEvent *k )
{
  int Taste;

  Taste = k->key();

	if( Taste == Key_Space )
  {
    emit SpacePressed();
    return;
  };
	if( Taste == Key_Up ) {
		cursorup = true;
		lastupdown = -1;
	};
	if( Taste == Key_Down ) {
		cursordown = true;
		lastupdown = 1;
	};
	if( Taste == Key_Left ) {
		cursorleft = true;
		lastleftright = -1;
  };
	if( Taste == Key_Right ) {
		cursorright = true;
		lastleftright = 1;
	};

	CalcCursorVector();
}


void MyWidget::keyReleaseEvent( QKeyEvent *k )
{
  int Taste;

  Taste = k->key();

  if( Taste == Key_Space ) return;

  if( Taste == Key_Up ) cursorup = false;
  if( Taste == Key_Down ) cursordown = false;
  if( Taste == Key_Left ) cursorleft = false;
  if( Taste == Key_Right ) cursorright = false;

  CalcCursorVector();
}


void MyWidget::CalcCursorVector()
{
  static QPoint CursorVector = QPoint(0,0);

  if( cursorup && !cursordown ) CursorVector.setY(-1);
  if( cursordown && !cursorup ) CursorVector.setY(1);
  if( cursordown && cursorup ) CursorVector.setY(lastupdown);
  if( !cursorup && !cursordown ) CursorVector.setY(0);
  if( cursorleft && !cursorright ) CursorVector.setX(-1);
  if( cursorright && !cursorleft ) CursorVector.setX(1);
  if( cursorright && cursorleft ) CursorVector.setX(lastleftright);
  if( !cursorleft && !cursorright ) CursorVector.setX(0);

  emit CursorVectorChange( CursorVector );
    
}


long MyWidget::ReadRecordFile()
{
  // Liest die Datei mit dem Zeitrekord ein
  Q_INT32 a;
  QDir FilePath;
  QString StrFileName;
  QString StrFilePath;
  QString StrNumber;
     
  FilePath = RecordFilePath();
  if ( FilePath.isRoot() ) return 17999999;

  StrFileName = QString("labyrinth.record");
  StrFilePath = FilePath.path();

  // qDebug( " Pfad zum lesen ist: " + StrFilePath );
  // qDebug( " Versuche Datei zu lesen. ");

  QFile f( StrFilePath+QString("/")+StrFileName );    // Kopiert aus QT Dokumentation
  if ( f.open( IO_ReadOnly ) ) {                      // open file for reading
    QDataStream s( &f );                              // serialize using f
    s >> a;                                           // Stream s nach a ... "the answer is" and 42
    // qDebug( " Datei oeffnen ( lesen ) erfolgreich. ");
    StrNumber.setNum(a);
    // qDebug( " Rekord ist " + StrNumber );
  }
  else {
    // qDebug( " Lesen nicht erfolgreich. ");
    WriteRecordFile( 17999999 , FilePath);
    a = 17999999;
  };

  return a;
}  


void MyWidget::WriteRecordFile( long Rekord )
{
  // Schreibt die Datei mit dem Zeitrekord, ermittelt zunchst den Pfad  
  QDir FilePath;
  
  FilePath = RecordFilePath();
  if ( FilePath.isRoot() ) return;
      
  WriteRecordFile( Rekord , FilePath );
  
}


void MyWidget::WriteRecordFile( long Rekord , QDir FilePath )
{
  // Schreibt die Datei mit dem Zeitrekord
  QString StrFileName;
  QString StrFilePath;

  if ( FilePath.isRoot() ) return;

  StrFilePath = FilePath.path();
  StrFileName = QString("labyrinth.record");

  // qDebug( " Pfad zum schreiben ist: " + StrFilePath );
  // qDebug( " Versuche Datei zu schreiben. ");


  QFile f( StrFilePath+QString("/")+StrFileName );  // kopiert aus der QT Doku
  if ( f.open( IO_WriteOnly ) ) {                   // open file for writing
    QDataStream s( &f );                            // serialize using f
    s << (Q_INT32)Rekord;                           // serialize integer
    // qDebug( " Datei oeffnen ( schreiben ) erfolgreich. ");
  }
  else {
    // qDebug( " Schreiben nicht erfolgreich. ");
  };
  
}


QDir MyWidget::RecordFilePath()
{
  // Ermittelt den Pfad fr den Zeitrekord. Zur Not wird das Verzeichnis erstellt.
  static bool bekannt = false;
  static QDir FilePath;

  QFile RecordFile;
  
  QString SDCard;
  QString CFCard;
  QString RestPfad;
  QString FileName;

  if ( bekannt ) return FilePath;

  // Suche den Filepath
  // qDebug( " Suche Pfad fuer RecordFile. ");
  SDCard = QString("/mnt/card");
  CFCard = QString("/mnt/cf");
  RestPfad = QString(QPEApplication::qpeDir()+"/share/games/labyrinth");  //  PC: /opt/Qtopia/... Z: /opt/QtPalmtop/...
  FileName = QString("labyrinth.record");
 
  
  if ( FilePath.exists(QString(SDCard+RestPfad)) ) {
    FilePath.setPath(QString(SDCard+RestPfad));
    bekannt = true;
    // qDebug( " Pfad auf SC-Card gesetzt. ");
    return FilePath;
  };  
  if ( FilePath.exists(QString(CFCard+RestPfad)) ) {
    FilePath.setPath(QString(CFCard+RestPfad));
    bekannt = true;
    // qDebug( " Pfad auf CF-Card gesetzt. ");
    return FilePath;
  };
  if ( FilePath.exists(RestPfad) ) {
    FilePath.setPath(RestPfad);
    bekannt = true;
    // qDebug( " Pfad auf RAM gesetzt. ");
    return FilePath;
  };

  if ( FilePath.mkdir(RestPfad) ) {
    FilePath.setPath(RestPfad);
    // qDebug( " Pfad im RAM erzeugt. ");
    if ( !RecordFile.exists(RestPfad+QString("/")+FileName) ) {
      // qDebug( " Versuche Datei zu erzeugen. ");
      WriteRecordFile( 17999999 , FilePath);
    };
    bekannt = true;
    return FilePath;
  }  
  else {
    FilePath.setPath("/");
    bekannt = true;
    // qDebug( " Konnte Verzeichnis weder finden noch erzeugen! ");
         
    return FilePath;    
  };
  
}  




