
/**********************************************************************
 *           labyrinth.h  -  description
 *           Project: "Labyrinth"
 *
 *           (C) 2003 by Helge Plehn
 *           29.03.2003
 *                                                                         
 *   This program is free software; you can redistribute it and/or modify  
 *   it under the terms of the GNU General Public License as published by 
 *   the Free Software Foundation; either version 2 of the License, or     
 *   (at your option) any later version.                               
 *                                                                    
**********************************************************************/

#ifndef MYWIDGET_H
#define MYWIDGET_H

#include <qwidget.h>
#include <qevent.h>
#include <qdatastream.h>
#include <qfile.h>
#include <qdir.h> 
#include <qlcdnumber.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qpushbutton.h>
#include <qmessagebox.h>
#include <qpoint.h>
#include <qarray.h>
#include <stdio.h>
#include <qfont.h>
#include <qstring.h>

#include <qglobal.h> // only for Debugging

#include "gamefield.h"
#include "showvector.h"
#include "myline.h"
#include "myblock.h"


class MyWidget : public QWidget
{
Q_OBJECT
public:
  MyWidget( QWidget *parent=0, const char *name=0, WFlags fl = 0 );
  static QString appName() { return QString::fromLatin1("labyrinth"); }

  //unsigned long lsqrt (unsigned long x);

public slots:
  
  void help();
  void about();
  void StartGame();
  void SetTime( long time );

  void PushStartPause();

  void setgamemode( int Modus );

  void ChangeCheatMode();
  //void ChangeFriction();
  void DeleteRecord();
  
protected:
  void keyPressEvent( QKeyEvent *k );
  void keyReleaseEvent( QKeyEvent *k );

signals:

  void EmitStartGame();
  void EmitPauseGame();
  void SpacePressed();
  void CursorVectorChange( QPoint CursorVector );
  
  void SignalChangeCheatMode(bool);
  //void SignalChangeFriction(bool);
  
private:
  void CalcCursorVector();
  
  int gamemode;
  long gametime;
  long timerecord;


  QString ConvertInt( long Nummer );
  QString ConvertTime( long LongTime );

  long MyWidget::ReadRecordFile();
  void MyWidget::WriteRecordFile( long Rekord );
  void MyWidget::WriteRecordFile( long Rekord , QDir FilePath );
  QDir MyWidget::RecordFilePath();

  //Status
	QLCDNumber *LCDtime;
	QLCDNumber *LCDtimerecord;

	//Child-Widgets
	GameField *gameField;
  ShowVector *showVector;

	//Menu
  QMenuBar *menu;
  QPopupMenu *options;
  
  //int FrictionID;
  //bool Friction;
  int CheatingID;
  bool CheatMode;
  
  //PushButton
  QPushButton *StartPause;  

	//Labyrinth tilting
  bool cursorup;
  bool cursordown;
  bool cursorleft;
  bool cursorright; 
  int lastupdown;
  int lastleftright;

};




#endif

