/**********************************************************************
 *           myblock.cpp  -  description
 *           Project: "Labyrinth"
 *
 *           (C) 2003 by Helge Plehn
 *           29.03.2003
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
**********************************************************************/

#include "myblock.h"

MyBlock::MyBlock( QPoint P1, QPoint P2 )
{
  // Constructer
  Punkt1 = P1;
  Punkt2 = P2;
  Initialisierung();
} 

MyBlock::MyBlock( )
{
  // Alternativer Constructer
  Punkt1 = QPoint(0,0);
  Punkt2 = QPoint(0,0);
  Initialisierung();
}
 
void MyBlock::SetP1( QPoint P1 )
{
  Punkt1 = P1;
  Initialisierung();
 }
 
void MyBlock::SetP2( QPoint P2 )
{
  Punkt2 = P2;
  Initialisierung();
}

QPoint MyBlock::GetP1( )
{ return Punkt1; }

QPoint MyBlock::GetP2( )
{ return Punkt2; }

int MyBlock::MinX()
{ return MinimumX; }

int MyBlock::MaxX()
{  return MaximumX; }

int MyBlock::MinY()
{ return MinimumY; }

int MyBlock::MaxY()
{  return MaximumY; }

void MyBlock::Initialisierung()
{
  if( Punkt1.x() < Punkt2.x() ) {
    MinimumX = Punkt1.x()-256;
    MaximumX = Punkt2.x()+256;
  }
  else {
    MinimumX = Punkt2.x()-256;
    MaximumX = Punkt1.x()+256;
  };
  if( Punkt1.y() < Punkt2.y() ) {
    MinimumY = Punkt1.y()-256;
    MaximumY = Punkt2.y()+256;
  }
  else {
    MinimumY = Punkt2.y()-256;
    MaximumY = Punkt1.y()+256;
  };
  
  AussenEcken[0] = QPoint(MinimumX,MinimumY); // links oben
  AussenEcken[1] = QPoint(MaximumX,MinimumY); // rechts oben
  AussenEcken[2] = QPoint(MaximumX,MaximumY); // rechts unten
  AussenEcken[3] = QPoint(MinimumX,MaximumY); // links unten

  InnenEcken[0] = AussenEcken[0] - QPoint(-256,-256); // links oben
  InnenEcken[1] = AussenEcken[0] - QPoint(256,-256);  // rechts oben
  InnenEcken[2] = AussenEcken[0] - QPoint(256,256);   // rechts unten
  InnenEcken[3] = AussenEcken[0] - QPoint(-256,256);  // links unten

  Begrenzungslinien[0] = MyLine(QPoint(MinimumX+256,MinimumY),QPoint(MaximumX-256,MinimumY)); // Oberkante
  Begrenzungslinien[1] = MyLine(QPoint(MaximumX,MinimumY+256),QPoint(MaximumX,MaximumY-256)); // rechte Kante
  Begrenzungslinien[2] = MyLine(QPoint(MinimumX+256,MaximumY),QPoint(MaximumX-256,MaximumY)); // Unterante
  Begrenzungslinien[3] = MyLine(QPoint(MinimumX,MinimumY+256),QPoint(MinimumX,MaximumY-256)); // linke Kante

  EckKanten[0] = MyLine(QPoint(MinimumX,MinimumY+256),QPoint(MinimumX+256,MinimumY)); // links oben
  EckKanten[1] = MyLine(QPoint(MaximumX,MinimumY+256),QPoint(MaximumX-256,MinimumY)); // rechts oben
  EckKanten[2] = MyLine(QPoint(MaximumX-256,MaximumY),QPoint(MaximumX,MaximumY-256)); // rechts unten
  EckKanten[3] = MyLine(QPoint(MinimumX+256,MaximumY),QPoint(MinimumX,MaximumY-256)); // links unten

  for ( int i = 0 ; i < 4 ; i++ )
  {
     EckKantenRotiert[i] = MyLine( EckKanten[i].Rotate45( EckKanten[i].GetP1() , 1), EckKanten[i].Rotate45( EckKanten[i].GetP2() , 1) );

     // Rundungsfehler ausgleichen  (Abweichungen <=3 werden gleichgesetzt!)
     if ( QABS(EckKantenRotiert[i].DeltaX())<=3 && QABS(EckKantenRotiert[i].DeltaX())>0 ) {
       EckKantenRotiert[i].SetP1( QPoint( ( EckKantenRotiert[i].MaxX()+EckKantenRotiert[i].MinX() )/2 , EckKantenRotiert[i].GetP1().y() ) );
       EckKantenRotiert[i].SetP2( QPoint(  EckKantenRotiert[i].GetP1().x() , EckKantenRotiert[i].GetP2().y() ) );
       //qDebug( " X-Position korrigiert! " );
     };
     if ( QABS(EckKantenRotiert[i].DeltaY())<=3 && QABS(EckKantenRotiert[i].DeltaY())>0 ) {
       EckKantenRotiert[i].SetP1( QPoint( EckKantenRotiert[i].GetP1().x() , ( EckKantenRotiert[i].MaxY()+EckKantenRotiert[i].MinY() )/2  ) );
       EckKantenRotiert[i].SetP2( QPoint( EckKantenRotiert[i].GetP2().x() , EckKantenRotiert[i].GetP1().y() ) );
       //qDebug( " Y-Position korrigiert! " );
     };

     //if ( EckKantenRotiert[i].DeltaX() != 0 && EckKantenRotiert[i].DeltaY() != 0 ) qDebug( " Rotierte Eckkante nicht orthogonal! " );
                                                                
  };

  
}


bool MyBlock::IsInside(QPoint Position)
{
   // berprft ob sich die Position innerhalb des Hindernisses befindet. Die Begrenzungslinie zhlt als ausserhalb!
   QPoint Difference;
   
   if ( Position.x() <= MinimumX ||  Position.x() >= MaximumX || Position.y() <= MinimumY ||  Position.y() >= MaximumY ) return false;  // Position ist ausserhalb dieses Obstacles
   for ( int i = 0 ; i < 4 ; i++ )
   {
      Difference = Position - AussenEcken[i];
      if ( QABS(Difference.x())+QABS(Difference.y()) <= 256 ) return false;  // Position ist innerhalb des durch die Aussenecken aufgepannten Rechteckes, aber zu nahe an einer der Aussenecken, so dass sie als ausserhalb angesehen wird.
   };
   return true;
}



MyLine MyBlock::Schnittpunkt(QPoint Position, QPoint OldPosition, QPoint PosRot, QPoint OldPosRot)
{
   //QPoint Difference;
   QPoint TestSchnittpunkt;
   QPoint MinSchnittpunkt = QPoint(0,0);
   QPoint SpareSchnittpunkt = QPoint(0,0);
   MyLine MotionVector = MyLine(OldPosition,Position);
   MyLine MotionVectorRotiert = MyLine(OldPosRot, PosRot);

   //QPoint Difference; // Debug Print!
   
   unsigned long MinAbstandSchnittpunkt = 131072;
   unsigned long SpareAbstandSchnittpunkt = 524288;
   unsigned long AbstandSchnittpunkt;
   int Kante=1; // 1 = Kante, 0 = Ecke
   int Index=0; //
   int SpareKante=1; // 1 = Kante, 0 = Ecke
   int SpareIndex=0; //

   if ( !IsInside( Position ) ) return MyLine(QPoint(0,0),QPoint(0,0));  // Position ist ausserhalb dieses Hindernisses
   
   for ( int i = 0 ; i < 4 ; i++ )
   {
      if ( (Begrenzungslinien[i].MaxX() >= MotionVector.MinX()) && (Begrenzungslinien[i].MinX() <= MotionVector.MaxX()) && (Begrenzungslinien[i].MaxY() >= MotionVector.MinY()) && (Begrenzungslinien[i].MinY() <= MotionVector.MaxY()) ) {
         TestSchnittpunkt = Begrenzungslinien[i].Intersection( MotionVector );
         AbstandSchnittpunkt = PunktAbstand( TestSchnittpunkt , OldPosition );
         if ( Begrenzungslinien[i].IsOnLine(TestSchnittpunkt) ) {
            if ( AbstandSchnittpunkt < MinAbstandSchnittpunkt ) {
               MinSchnittpunkt = TestSchnittpunkt;
               MinAbstandSchnittpunkt = AbstandSchnittpunkt;
               Kante = 1;
               Index = i;
            };   
         }
         else {
            if ( AbstandSchnittpunkt < SpareAbstandSchnittpunkt ) {
               SpareSchnittpunkt = TestSchnittpunkt;
               SpareAbstandSchnittpunkt = AbstandSchnittpunkt;
               SpareKante = 1;
               SpareIndex = i;
            };
         };
      };
   };

   
   for ( int i = 0 ; i < 4 ; i++ )
   {
      if ( (EckKanten[i].MaxX() >= MotionVector.MinX()) && (EckKanten[i].MinX() <= MotionVector.MaxX()) && (EckKanten[i].MaxY() >= MotionVector.MinY()) && (EckKanten[i].MinY() <= MotionVector.MaxY()) ) {
         TestSchnittpunkt = EckKantenRotiert[i].Intersection( MotionVectorRotiert );
         //qDebug( " Schnittpunkt geprft. " );            
         AbstandSchnittpunkt = PunktAbstand( TestSchnittpunkt , OldPosRot );
         if ( EckKantenRotiert[i].IsOnLine(TestSchnittpunkt) ) {
            //qDebug( " Schnittpunkt gefunden. " );
            if ( AbstandSchnittpunkt < MinAbstandSchnittpunkt ) { 
              MinSchnittpunkt = EckKantenRotiert[i].Rotate45(TestSchnittpunkt,-1); // Zurckdrehen
              MinAbstandSchnittpunkt = AbstandSchnittpunkt;
              Kante = 0;
              Index = i;            
            };
         }
         else {
            if ( AbstandSchnittpunkt < SpareAbstandSchnittpunkt ) {
               SpareSchnittpunkt = EckKantenRotiert[i].Rotate45(TestSchnittpunkt,-1);
               SpareAbstandSchnittpunkt = AbstandSchnittpunkt;
               SpareKante = 0;
               SpareIndex = i;
            };
         };
      };   
   };  

   // Falls MinSchnittpunkt == QPoint(0,0): Fehler, kein Schnittpunkt gefunden, obwohl Position innerhalb des Hindernisses!

   if ( MinSchnittpunkt == QPoint(0,0) && SpareSchnittpunkt !=  QPoint(0,0) ) {
     MinSchnittpunkt = SpareSchnittpunkt; 
     Kante = SpareKante;
     Index = SpareIndex;
     // qDebug( " Fehler bei Schnittpunkterkennung, Ersatzschnittpunkt wird verwendet. " );          
   };
   
   /*if ( MinSchnittpunkt == QPoint(0,0) && SpareSchnittpunkt ==  QPoint(0,0) ) {
      qDebug( " Fehler bei Schnittpunkterkennung! " );
      qDebug( " Position: " + StringFromNumber(Position) + " , OldPosition: " + StringFromNumber(OldPosition));
      qDebug( " PosRot:   " + StringFromNumber(PosRot) +   " , OldPosRot:   " + StringFromNumber(OldPosRot));
      for ( int i = 0 ; i < 4 ; i++ )
      {
         qDebug( " Eckkante[" + StringFromNumber(i) +   "] : " + StringFromNumber(EckKanten[i]));
         Difference=AussenEcken[i]-Position;
         qDebug( " Aussenecke: " + StringFromNumber(AussenEcken[i]) + " , Differenz: " + StringFromNumber(Difference) + " Abstand: " + StringFromNumber(abs(Difference.x())+abs(Difference.y())) );
         Difference=InnenEcken[i]-Position;
         qDebug( " Innenecke:  " + StringFromNumber(InnenEcken[i]) + " , Differenz: " + StringFromNumber(Difference) + " Abstand: " + StringFromNumber(abs(Difference.x())+abs(Difference.y())) );
      };
   };*/

   return MyLine(MinSchnittpunkt, QPoint(Kante, Index));
   
}




QPoint MyBlock::Reflection(QPoint Position, QPoint Schnittpunkt, int Kante, int Index)
{
   QPoint NewPosition;
   int Entscheider;

   NewPosition = Position;
   Entscheider = Index + 4*Kante;

   switch ( Entscheider ) {
     case 0: // links oben
        NewPosition.setX( 2 * Schnittpunkt.x() - Position.x() -1 );
        NewPosition.setY( 2 * Schnittpunkt.y() - Position.y() -1 );
     break;
     case 1: // rechts oben
        NewPosition.setX( 2 * Schnittpunkt.x() - Position.x() +1 );
        NewPosition.setY( 2 * Schnittpunkt.y() - Position.y() -1 );
     break;
     case 2: // rechts unten
        NewPosition.setX( 2 * Schnittpunkt.x() - Position.x() +1 );
        NewPosition.setY( 2 * Schnittpunkt.y() - Position.y() +1 );
     break;
     case 3: // links unten
        NewPosition.setX( 2 * Schnittpunkt.x() - Position.x() -1 );
        NewPosition.setY( 2 * Schnittpunkt.y() - Position.y() +1 );
     break;
     case 4: // Oberkante
        NewPosition.setY( 2 * Schnittpunkt.y() - Position.y() -1 );
     break;
     case 5: // rechte Kante
        NewPosition.setX( 2 * Schnittpunkt.x() - Position.x() +1 );
     break;
     case 6: // Unterkante
        NewPosition.setY( 2 * Schnittpunkt.y() - Position.y() +1 );
     break;
     case 7: // linke Kante
        NewPosition.setX( 2 * Schnittpunkt.x() - Position.x() -1 );
     break;
   };

          
   return NewPosition;
}



QPoint MyBlock::SpeedReflection(QPoint Speed, int Kante, int Index, int Limit)
{
   QPoint NewSpeed= Speed;
   MyLine NeedObj = MyLine();

   if ( Kante == 1 ) {
     // Kanten
     NewSpeed = Speed;
   }
   else {
     // Ecken
     NewSpeed = NeedObj.Rotate45(Speed,1);
   };
   
   switch ( Index ) {
     case 0: // Oberkante oder links oben
        NewSpeed.setY(- QABS(Speed.y()) );        
     break;
     case 1: // rechte Kante oder rechts oben
        NewSpeed.setX( QABS(Speed.x()) );
     break;
     case 2: // Unterkante oder rechts unten
        NewSpeed.setY( QABS(Speed.y()) );
     break;
     case 3: // linke Kante oder links unten
        NewSpeed.setX(-QABS(Speed.x()) );
     break;
   };

   NewSpeed = SpeedDamping( NewSpeed, Index, Limit );
   
   if ( Kante == 1 ) {
     // Kanten
     return NewSpeed;
   }
   else {
     // Ecken
     return NeedObj.Rotate45(NewSpeed,-1);
   };

}

QPoint MyBlock::SpeedDamping(QPoint Speed, int Index, int Limit)
{
  QPoint NewSpeed = Speed;

  switch ( Index ) {
     case 0: // oben oder links oben
     case 2: // unten oder rechts unten
       if ( QABS(Speed.y()) > Limit ) {
          NewSpeed.setY( Speed.y() /2 );
       }
       else {
         NewSpeed.setY(0);
       };    
     break;
     case 1: // rechts oder rechts oben
     case 3: // links oder links unten
       if ( QABS(Speed.x()) > Limit ) {
          NewSpeed.setX( Speed.x() /2 );
       }
       else {
         NewSpeed.setX(0);
       };
     break;
  };   

  return NewSpeed;
  
}




unsigned long MyBlock::PunktAbstand( QPoint Punkt1, QPoint Punkt2 )
{
  QPoint Difference;
  unsigned long Abstand;
  unsigned long AbsDeltaX;
  unsigned long AbsDeltaY;

  Difference = Punkt2 - Punkt1;

  AbsDeltaX = static_cast<unsigned long>(QABS(Difference.x()));
  AbsDeltaY = static_cast<unsigned long>(QABS(Difference.y()));
  Abstand = AbsDeltaX*AbsDeltaX+AbsDeltaY*AbsDeltaY;

  return Abstand;
}

int MyBlock::Signum( int Argument )
{
  return (Argument < 0) ? -1 : (0 < Argument) ? 1 : 0 ;
}


/* // Only for debugging

QString MyBlock::StringFromNumber( long Number )
{
  QString Antwort;
  Antwort.setNum(Number);
  return Antwort;
}  

QString MyBlock::StringFromNumber( int Number )
{
  QString Antwort;
  Antwort.setNum(Number);
  return Antwort;
}

QString MyBlock::StringFromNumber( unsigned long int Number )
{
  QString Antwort;
  Antwort.setNum(Number);
  return Antwort;
}

QString MyBlock::StringFromNumber( QPoint Punkt )
{
  QString xString;
  QString yString;
  xString.setNum(Punkt.x());
  yString.setNum(Punkt.y());
  return QString("("+xString+";"+yString+")");
}

QString MyBlock::StringFromNumber( MyLine Line )
{
  QString Antwort;
  Antwort = QString("{"+StringFromNumber(Line.GetP1())+";"+StringFromNumber(Line.GetP2())+"}");
  return Antwort;
}

*/


